#include <so_5/all.hpp>

using namespace std::chrono_literals;

// Type of message to be sent to file_writer_thread.
struct write_data {
   const std::string file_name_;
};

// Data "acquisition" thread.
void meter_reader_thread(
      // A channel for timer signals.
      so_5::mchain_t timer_ch,
      // A channel for communication with file_writer_thread.
      so_5::mchain_t file_write_ch) {

   // A type for periodic timer signal.
   struct acquisition_turn : public so_5::signal_t {};

   // Just a counter for generation of new file names.
   int ordinal = 0;

   // Start periodic timer.
   auto timer = so_5::send_periodic<acquisition_turn>(timer_ch, 0ms, 750ms);

   // Read all messages from the channel until the channel will be closed.
   receive(from(timer_ch),
         // A handler for acquisition_turn signal.
         [&](so_5::mhood_t<acquisition_turn>) {
            // Simulate data acquisition.
            std::cout << "meter read started" << std::endl;
            std::this_thread::sleep_for(50ms);
            std::cout << "meter read finished" << std::endl;

            // Send a command to write a new file.
            so_5::send<write_data>(file_write_ch,
                  "data_" + std::to_string(ordinal) + ".dat");
            ++ordinal;
         });
}

// A thread for "writing" data to files.
void file_writer_thread(
      // A channel for incoming commands.
      so_5::mchain_t file_write_ch) {
   // Read all messages from the channel until the channel will be closed.
   receive(from(file_write_ch),
         // A handler for write_data messages.
         [&](so_5::mhood_t<write_data> cmd) {
            // Simulate data storing.
            std::cout << cmd->file_name_ << ": write started" << std::endl;
            std::this_thread::sleep_for(350ms);
            std::cout << cmd->file_name_ << ": write finished" << std::endl;
         });
}

int main() {
   // Launch SObjectizer.
   so_5::wrapped_env_t sobj;

   // Objects for thread created now...
   std::thread meter_reader, file_writer;
   // ...just to create auto_joiner object before auto_closer object.
   auto joiner = so_5::auto_join(meter_reader, file_writer);

   // Create channels for our threds.
   auto timer_ch = so_5::create_mchain(sobj);
   auto writer_ch = so_5::create_mchain(sobj);
   // Channels must be automatically closed on exit.
   auto closer = so_5::auto_close_drop_content(timer_ch, writer_ch);

   // Now we can start our threads.
   meter_reader = std::thread(meter_reader_thread, timer_ch, writer_ch);
   file_writer = std::thread(file_writer_thread, writer_ch);

   // We should work until user enters 'exit' or closes the input stream.
   std::cout << "Type 'exit' to quit:" << std::endl;

   std::string cmd;
   while(std::getline(std::cin, cmd)) {
      if("exit" == cmd)
         break;
      else
         std::cout << "Type 'exit' to quit" << std::endl;
   }

   // Just returns from main(). All channels will be closed automatically
   // (because of closer object) and all threads will be joined automatically
   // (because of joiner object).
   return 0;
}

