#include <so_5/all.hpp>

#include <random>

using namespace std::chrono_literals;

// Type of message to be sent to file_writer_thread.
struct write_data {
   const std::string file_name_;
};

// A command to decrease data acquisition period.
struct dec_read_period : public so_5::signal_t {};

// A command to increase data acquisition period.
struct inc_read_period : public so_5::signal_t {};

// Just a helper function for conversion from arbitrary duration into milliseconds.
template<typename R, typename P>
std::chrono::milliseconds to_ms(std::chrono::duration<R, P> v) {
   return std::chrono::duration_cast<std::chrono::milliseconds>(v);
}

// Data "acquisition" thread.
void meter_reader_thread(
      // A command channel.
      so_5::mchain_t control_ch,
      // A channel for timer signals.
      so_5::mchain_t timer_ch,
      // A channel for communication with file_writer_thread.
      so_5::mchain_t file_write_ch) {

   // Range for allowed acquisition period.
   const auto period_left_range = to_ms(50ms);
   const auto period_right_range = to_ms(2s);

   // Current acquisition period.
   auto current_period = to_ms(300ms);

   // A type for periodic timer signal.
   struct acquisition_turn : public so_5::signal_t {};

   // Just a counter for generation of new file names.
   int ordinal = 0;

   // Send the first timer signal for itself.
   so_5::send<acquisition_turn>(timer_ch);

   // Read all messages from the channel until all channels will be closed.
   so_5::select(so_5::from_all(),
      // Handlers for timer channel.
      case_(timer_ch,
         // A handler for acquisition_turn signal.
         [&](so_5::mhood_t<acquisition_turn>) {
            // We should know how much time was spent.
            // Take the first timepoint.
            const auto started_at = std::chrono::steady_clock::now();

            // Simulate data acquisition.
            std::cout << "meter read started" << std::endl;
            std::this_thread::sleep_for(50ms);
            std::cout << "meter read finished" << std::endl;

            // Send a command to write a new file.
            so_5::send<write_data>(file_write_ch,
                  "data_" + std::to_string(ordinal) + ".dat");
            ++ordinal;

            // Now we can calculate time spent.
            const auto duration = std::chrono::steady_clock::now() - started_at;
            // If procedure lasts too long then the next acquisition
            // must be scheduled immediately.
            if(duration >= current_period) {
               std::cout << "period=" << current_period.count()
                     << "ms, no sleep" << std::endl;
               so_5::send<acquisition_turn>(timer_ch);
            }
            else {
               // Otherwise we can sleep for some time.
               const auto sleep_time = to_ms(current_period - duration);
               std::cout << "period=" << current_period.count() << "ms, sleep="
                     << sleep_time.count() << "ms" << std::endl;
               so_5::send_delayed<acquisition_turn>(timer_ch,
                     current_period - duration);
            }
         }),
      // Handlers for commands from control channel.
      case_(control_ch,
         [&](so_5::mhood_t<inc_read_period>) {
            current_period = std::min(to_ms(current_period * 1.5), period_right_range);
         },
         [&](so_5::mhood_t<dec_read_period>) {
            current_period = std::max(to_ms(current_period / 1.5), period_left_range);
         })
   );
}

// A thread for "writing" data to files.
void file_writer_thread(
      // A channel for incoming commands.
      so_5::mchain_t file_write_ch) {
   // Helpers for random values generation.
   std::mt19937 rd_gen{std::random_device{}()};
   // Pauses will be generated from a range [295ms, 1s].
   std::uniform_int_distribution<int> rd_dist{295, 1000};

   // Read all messages from the channel until the channel will be closed.
   receive(from(file_write_ch),
         // A handler for write_data messages.
         [&](so_5::mhood_t<write_data> cmd) {
            // Generate random pause for simulation.
            const auto pause = rd_dist(rd_gen);
            // Simulate data storing.
            std::cout << cmd->file_name_ << ": write started (pause:"
                  << pause << "ms)" << std::endl;
            std::this_thread::sleep_for(std::chrono::milliseconds{pause});
            std::cout << cmd->file_name_ << ": write finished" << std::endl;
         });
}

int main() {
   // Launch SObjectizer.
   so_5::wrapped_env_t sobj;

   // Objects for thread created now...
   std::thread meter_reader, file_writer;
   // ...just to create auto_joiner object before auto_closer object.
   auto joiner = so_5::auto_join(meter_reader, file_writer);

   // Create channels for our threds.
   //
   // Control channel for meter_reader_thread. Without any restrictions.
   auto control_ch = so_5::create_mchain(sobj);
   // A timer channel for meter_reader_thread.
   auto timer_ch = so_5::create_mchain(control_ch->environment(),
         // Just one message can wait in the channel.
         1,
         // Memory will be preallocated.
         so_5::mchain_props::memory_usage_t::preallocated,
         // Newest message will be ignored if the channel is full.
         so_5::mchain_props::overflow_reaction_t::drop_newest);
   // A channel with "acquired" data will be limited in size,
   // with a waiting on full channel and with removing the oldest message
   // policy.
   auto writer_ch = so_5::create_mchain(sobj,
         // Wait on full channel no more than 300ms.
         300ms,
         // No more than 2 messages in the channel.
         2,
         // Memory will be preallocated.
         so_5::mchain_props::memory_usage_t::preallocated,
         // The oldest message should be removed on overflow.
         so_5::mchain_props::overflow_reaction_t::remove_oldest);
   // Channels must be automatically closed on exit.
   auto closer = so_5::auto_close_drop_content(control_ch, timer_ch, writer_ch);

   // Now we can start our threads.
   meter_reader = std::thread(meter_reader_thread, control_ch, timer_ch, writer_ch);
   file_writer = std::thread(file_writer_thread, writer_ch);

   // We should work until user enters 'exit' or closes the input stream.
   bool stop_execution = false;
   while(!stop_execution) {
      std::cout << "Type 'exit' to quit, 'inc' or 'dec':" << std::endl;

      std::string cmd;
      if(std::getline(std::cin, cmd)) {
         if("exit" == cmd)
            stop_execution = true;
         else if("inc" == cmd)
            so_5::send<inc_read_period>(control_ch);
         else if("dec" == cmd)
            so_5::send<dec_read_period>(control_ch);
      }
      else
         stop_execution = true;
   }

   // Just returns from main(). All channels will be closed automatically
   // (because of closer object) and all threads will be joined automatically
   // (because of joiner object).
   return 0;
}

